// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Bandwidth Divider                                                ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


import TransportInfo;


/**
  * This class contains a static method to devide up an upper bandwidth
  * limit between multiple clients.
  *
  * @short   Bandwidth Divider
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
public class BandwidthDivider
{
   // ###### Private data ###################################################
   private final static boolean PrintResults = false;   // Print all results

   private static long Rest;


   // ###### Constructor ####################################################
   public BandwidthDivider() {
   }


   // ###### Get minimum ####################################################
   /**
     * Get minimum of a and b:
     *
     * @param a Value 1.
     * @param b Value 2.
     * @return Minimum of a and b.
     */
   public static long min(long a, long b) {
      if(a < b)
         return(a);
      return(b);
   }


   // ###### Give bandwidth to stream #######################################
   private boolean giveBandwidth(long[] wanted,
                                 long[] bandwidth,
                                 int    layers) {
      // ====== Calculate total wanted and owned values =====================
      boolean changed = true;
      long totalWanted = wanted[0];
      long totalOwned  = bandwidth[0];
      for(int i = 1;i < TransportInfoLevel.MaxQualityLayers;i++) {
         totalWanted += wanted[i];
         totalOwned  += bandwidth[i];
      }

      // ====== Try to give wanted value to stream ==========================
      if(totalWanted <= Rest + totalOwned) {
         for(int i = 0;i < layers;i++) {
            bandwidth[i] = wanted[i];
         }
         Rest = Rest + totalOwned - totalWanted;
      }
      else {
         changed = false;
      }
      return(changed);
   }


   // ###### Divide up bandwidth ############################################
   /**
     * Divide up an upper bandwidth limit.
     *
     * @param available Available bandwidth to be divided up.
     * @param transportInfo Array of TransportInfos.
     * @param bandwidth Array [ClientNum][LayerNum] to store calculated bandwidths for TransportInfos.
     * @param count Number of array entries.
     * @return Rest bandwidth.
     */
   synchronized public long divideUp(long            available, 
                                     TransportInfo[] transportInfo,
                                     long[][]        bandwidth,
                                     int             count) {
      // ====== Initialize ==================================================
      int i;
      int[] level   = new int[count];
      int[] scale   = new int[count];
      long[] wanted = new long[TransportInfoLevel.MaxQualityLayers];
      for(i = 0;i < TransportInfoLevel.MaxQualityLayers;i++) {
         wanted[i]    = 0;
         for(int j = 0;j < count;j++) {
            bandwidth[j][i] = 0;
         }
      }
      for(i = 0;i < count;i++) {
         level[i] = 0;
         scale[i] = 0;
      }

      // ====== Try to give each client required minimum of bandwidth =======
      Rest = available;
      for(i = 0;i < count;i++) {
         for(int j = 0;j < TransportInfoLevel.MaxQualityLayers;j++)
            wanted[j] = 0;
         for(int j = 0;j < transportInfo[i].QualityLevel[0].QualityLayers;j++) {
            wanted[j] = transportInfo[i].QualityLevel[0].QualityLayer[j].BytesPerSecond;
         }

         giveBandwidth(wanted,bandwidth[i],transportInfo[i].QualityLevel[0].QualityLayers);
      }

      // ====== Print results ===============================================
      if(PrintResults == true) {
         System.out.println("Step 1: Rest=" + Rest);
         for(i = 0;i < count;i++) {
            System.out.println("#" + i + ":  " +
               "L0=" + bandwidth[i][0] + "   " +
               "L1=" + bandwidth[i][1] + "   " +
               "L2=" + bandwidth[i][2] + "   " +
               "level=" + level[i]);
         }
      }


      // ====== Divide up remaining bandwidth ===============================
      /* 
         First, go up to wanted value.
         Second, go up to maximum value.
      */    
      long[][][] bandwidthLimit = new long[2][count][TransportInfoLevel.MaxQualityLayers];
      for(i = 0;i < count;i++) {
         for(int j = 0;j < transportInfo[i].QualityLayers;j++) {
            bandwidthLimit[0][i][j] = transportInfo[i].WantedBytesPerSecond[j];
            bandwidthLimit[1][i][j] = transportInfo[i].MaxBytesPerSecond[j];
         }
      }
      for(int dividerLevel = 0;dividerLevel < 2;dividerLevel++) {
         boolean changed;
         do {
            changed = false;
            for(i = 0;i < count;i++) {
               // ====== Check, if encoding wants more bandwidth ============
               boolean wantsMore = false;
               for(int j = 0;j < transportInfo[i].QualityLayers;j++) {
                  if(bandwidth[i][j] < bandwidthLimit[dividerLevel][i][j]) {
                     wantsMore = true;
                     break;
                  }
               }

               if(wantsMore) {
                  // ====== Scale for dynamic encodings =====================
                  if((transportInfo[i].Flags & TransportInfo.TIF_HasDynamic) == TransportInfo.TIF_HasDynamic) {
                     for(int j = 0;j < TransportInfoLevel.MaxQualityLayers;j++)
                        wanted[j] = 0;
                     for(int j = 0;j < transportInfo[i].QualityLevel[level[i]].QualityLayers;j++) {
                        wanted[j] = (transportInfo[i].WantedFramesPerSecond * 
                                       transportInfo[i].QualityLevel[level[i]].QualityLayer[j].FrameSize);

                        wanted[j] = wanted[j] +
                                    (long)(((double)scale[i] *
                                            (double)transportInfo[i].QualityLevel[level[i]].QualityLayer[j].FrameSize *
                                            (double)transportInfo[i].QualityLevel[level[i]].BytesPerSecondScale)
                                           / 65536.0);

                        if(wanted[j] > transportInfo[i].MaxBytesPerSecond[j])
                           wanted[j] = transportInfo[i].MaxBytesPerSecond[j];
                     }

                     changed = giveBandwidth(wanted,bandwidth[i],transportInfo[i].QualityLayers);
                     if(changed == true) {
                        scale[i]++;
                     }
                  }

                  // ====== Use next step for static encodings ==============
                  else {
                     int nextLevel = level[i] + 1;
                     if(nextLevel < transportInfo[i].QualityLevels) {
                        for(int j = 0;j < TransportInfoLevel.MaxQualityLayers;j++)
                           wanted[j] = 0;
                        for(int j = 0;j < transportInfo[i].QualityLevel[nextLevel].QualityLayers;j++) {
                           wanted[j] = transportInfo[i].QualityLevel[nextLevel].QualityLayer[j].BytesPerSecond;
                        }

                        changed = giveBandwidth(wanted,bandwidth[i],transportInfo[i].QualityLevel[nextLevel].QualityLayers);
                        if(changed == true) {
                           level[i] = nextLevel;
                        }
                     }
                  }
               }
            }
         } while(changed);
      }

      // ====== Print results ===============================================
      if(PrintResults == true) {
         System.out.println("Step 2: Rest=" + Rest);
         for(i = 0;i < count;i++) {
            System.out.println("#" + i + ":  " +
               "L0=" + bandwidth[i][0] + "   " +
               "L1=" + bandwidth[i][1] + "   " +
               "L2=" + bandwidth[i][2] + "   " +
               "level=" + level[i]);
         }
      }

      return(Rest);
   }
}
