// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### TransportInfo Layer                                              ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


/**
  * This class keeps information on a quality level's layers.
  *
  * @short   TransportInfo Layer
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
public class TransportInfoLayer
{
   // ====== Constructor ====================================================
   /**
     * Constructor.
     */
   public TransportInfoLayer() {
   }


   // ====== Reset ==========================================================
   /**
     * Reset.
     */
   public void reset() {
      BytesPerSecond   = 0;
      PacketsPerSecond = 0;
      FrameSize        = 0;
      MaxLossRate      = 255;
      Flags            = 0;
   }


   // ====== Print ==========================================================
   /**
     * Print.
     */
   public void print() {
      System.out.println("      BytesPerSecond   = " + BytesPerSecond);
      System.out.println("      PacketsPerSecond = " + PacketsPerSecond);
      System.out.println("      FrameSize        = " + FrameSize);
      System.out.println("      MaxLossRate      = " + MaxLossRate);
      System.out.println("      Flags            = " + Flags);
   }


   // ====== Layer's values =================================================
   /**
     * Layer's bytes per second at StartFramesPerSecond * FramesPerSecondScale
     * frames per second.
     */
   public long BytesPerSecond;

   /**
     * Layer's packets per second at StartFramesPerSecond * FramesPerSecondScale
     * frames per second.
     */
   public int PacketsPerSecond;

   /**
     * Layer's frame size.
     */
   public int FrameSize;

   /**
     * Layer's maximum acceptable loss rate in fraction of 255.
     */
   public int MaxLossRate;

   /**
     * Layer's flags.
     */
   public int Flags;

   /**
     * TransportInfoLayer flag: Static quality setting.
     */
   public static final int TILF_Static = 0;

   /**
     * TransportInfoLayer flag: Dynamic quality setting.
     */
   public static final int TILF_Dynamic = 1;
};
