// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Stream Source and Destination                                    ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#ifndef STREAMSRCDEST_H
#define STREAMSRCDEST_H


#include "system.h"
#include "internetaddress.h"


namespace Coral {


/**
  * This class is contains source and destination of a stream.
  *
  * @short   Stream Source and Destination
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class StreamSrcDest
{
   // ====== Constructor ====================================================
   public:
   /**
     * Constructor.
     */
   StreamSrcDest();


   // ====== Byte order translation =========================================
   /**
     * Translate byte order.
     */
   void translate();


   // ====== Reset ==========================================================
   /**
     * Reset.
     */
   void reset();


   // ====== Comparision operators ==========================================
   /**
     * == operator.
     */
   int operator==(const StreamSrcDest& ssd) const;

   /**
     * != operator.
     */
   int operator!=(const StreamSrcDest& ssd) const;


   // ====== Values =========================================================
   public:
   /**
     * Source address of the stream in portable address format.
     */           
   PortableAddress Source;

   /**
     * Destination address of the stream in portable address format.
     */
   PortableAddress Destination;

   /**
     * Flow label for IPv6 support.
     */            
   card32 FlowLabel;

   /**
     * Traffic class.
     */
   card8 TrafficClass;
   
   /**
     * Is this StreamSrcDest valid?
     */
   bool IsValid;
   
   /**
     * Padding for alignment.
     */
   card16 pad;  
};


/**
  * << operator.
  */
ostream& operator<<(ostream& os, const StreamSrcDest& ssd);


}


#endif
            