// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Dictionary Bag Template Implementation                           ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "bag.h"


// ###### Constructor #######################################################
template<class K, class T> DictionaryBag<K,T>::DictionaryBag()
{
   FirstNode.Prev = &FirstNode;
   FirstNode.Next = &FirstNode;
   Pointer        = &FirstNode;
   Count          = 0;
}


// ###### Destructor ########################################################
template<class K, class T> DictionaryBag<K,T>::~DictionaryBag()
{
   DictionaryBagNode<K,T>* node;
   DictionaryBagNode<K,T>* oldNode;
   node = FirstNode.Next;
   while(node != &FirstNode) {
      oldNode = node;      
      node = node->Next;
      delete oldNode;
   }
   FirstNode.Prev = NULL;
   FirstNode.Next = NULL;
}


// ###### Add element to head of bag ########################################
template<class K, class T> bool DictionaryBag<K,T>::addHead(K key, T element)
{
   DictionaryBagNode<K,T>* node = new DictionaryBagNode<K,T>;
   if(node != NULL) {
      node->Element    = element;
      node->Key        = key;
      node->Prev       = &FirstNode;
      node->Next       = FirstNode.Next;
      node->Next->Prev = node;
      FirstNode.Next   = node;   
      Pointer          = &FirstNode;
      Count++;
      return(true);
   }
   return(false);
}


// ###### Add element to tail of bag ########################################
template<class K, class T> bool DictionaryBag<K,T>::addTail(K key, T element)
{
   DictionaryBagNode<K,T>* node = new DictionaryBagNode<K,T>;
   if(node != NULL) {
      node->Element    = element;
      node->Key        = key;
      node->Prev       = FirstNode.Prev;
      node->Next       = &FirstNode;
      node->Prev->Next = node;
      FirstNode.Prev   = node;   
      Pointer          = &FirstNode;
      Count++;
      return(true);
   }
   return(false);
}


// ###### Remove element from bag ###########################################
template<class K, class T> void DictionaryBag<K,T>::remove(const K& key)
{
   DictionaryBagNode<K,T>* node = findNode(key);
   if(node) {
      removeNode(node);
      Pointer = &FirstNode;
   }
}


// ###### Remove all elements with given key from bag #######################
template<class K, class T> void DictionaryBag<K,T>::removeAll(const K& key)
{
   DictionaryBagNode<K,T>* node = findNode(key);
   while(node) {
      removeNode(node);
      Pointer = &FirstNode;
      node = findNode(key);
   }
}


// ###### Find DictionaryBagNode for element ##########################################
template<class K, class T> DictionaryBagNode<K,T>* DictionaryBag<K,T>::findNode(const K& key) const
{
   DictionaryBagNode<K,T>* node;
   node = FirstNode.Next;
   while(node != &FirstNode) {
      if(node->Key == key)
         return(node);
      node = node->Next;
   }      
   return(NULL);   
}


// ###### Remove element from bag ###########################################
template<class K, class T> void DictionaryBag<K,T>::remove(T element)
{
   DictionaryBagNode<K,T>* node = findNode(element);
   if(node) {
      removeNode(node);
      Pointer = &FirstNode;
   }
}


// ###### Find DictionaryBagNode for element ##########################################
template<class K, class T> DictionaryBagNode<K,T>* DictionaryBag<K,T>::findNode(const T& element) const
{
   DictionaryBagNode<K,T>* node;
   node = FirstNode.Next;
   while(node != &FirstNode) {
      if(node->Element == element)
         return(node);
      node = node->Next;
   }      
   return(NULL);   
}


// ###### Remove given node from bag ########################################
template<class K, class T> void DictionaryBag<K,T>::removeNode(DictionaryBagNode<K,T>* node)
{
   node->Prev->Next = node->Next;
   node->Next->Prev = node->Prev;
   Count--;
   delete node;
}
