// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Encoder Repository Interface                                     ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#ifndef ENCODERREPOSITORYINTERFACE_H
#define ENCODERREPOSITORYINTERFACE_H


#include "system.h"
#include "synchronizable.h"
#include "encoderinterface.h"


namespace Coral {


/**
  * This class is a repository for encoders.
  *
  * @short   Encoder Repository Interface
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class EncoderRepositoryInterface : virtual public EncoderInterface
{
   // ====== Select another encoder =========================================
   public: 
   /**
     * Select the encoder with the given TypeID to be the current encoder
     * of the repository.
     *
     * @param typeID Encoding's type ID.
     * @return true, if encoder for this TypeID was in the repository;
     * false otherwise.
     */
   virtual bool selectEncoderForTypeID(const card16 typeID) = 0;
   
   /**
     * Get EncoderInterface of the current encoder.
     *
     * @return Current encoder's EncoderInterface.     
     */
   virtual EncoderInterface* getCurrentEncoder() const = 0;
};


}


#endif
