// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Network Monitor Simulator Implementation                         ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "socket.h"
#include "internetaddress.h"
#include "congestionmonitor.h"
#include "networkmonitorinterface.h"
#include "tools.h"
#include "networkmonitorsimulator.h"


namespace Coral {


// ###### Constructor #######################################################
NetworkMonitorSimulator::NetworkMonitorSimulator(const card64 bandwidth)
{
   Report.reset();
   Bandwidth  = bandwidth;
   Congestion = 0.0;
}


// ###### Destructor ########################################################
NetworkMonitorSimulator::~NetworkMonitorSimulator()
{
}


// ###### Check, if NetworkMonitorSimulator is ready ########################
bool NetworkMonitorSimulator::ready() const
{
   return(true);
}


// ###### Lock monitor to access data #######################################
void NetworkMonitorSimulator::lock()
{
}


// ###### Unlock monitor ####################################################
void NetworkMonitorSimulator::unlock()
{
}


// ###### Reset report ######################################################
void NetworkMonitorSimulator::resetReport()
{
   Report.reset();
   TimeStamp = getMicroTime();
}


// ###### Get report ########################################################
void NetworkMonitorSimulator::getReport(NetworkMonitorReport* report)
{
   card64 now = getMicroTime();
   Report.Interval = now - TimeStamp;
   TimeStamp = now;

   Report.BytesRaw   = (card64)(Congestion * Bandwidth * (Report.Interval / 1000000.0));
   Report.PacketsRaw = (card32)(Report.BytesRaw / 512);
   memcpy(report,&Report,sizeof(Report));
}


// ###### Get report and reset ##############################################
void NetworkMonitorSimulator::moveReport(NetworkMonitorReport* report)
{
   getReport(report);
   resetReport();
}


}
