package bonn.infoIV.v3d2.rtp;

import java.util.Random;


public abstract class ControlChannel extends Thread implements Channel, BaseControlChannel
{
  protected DataChannel MyChannel_  = null; // the corresponding RTP-DataChannel
  protected RTCPPacket  RTCPPacket_ = null; // This packet is for receiving only
  /**
   * Running_ is true as long, as Channel Threads are not finished.
   * You have to set Running_ to 'false' to end a Channel-Thread properly.
   */
  protected boolean Running_     = true;

  /**
   * This is a reference to the Control-Object computing RTCP related data.
   * MyControl generates Network related information and Application related information
   */
  public    Control     MyControl_  = null;


  /**
   * This constructor instatiates variables RTCPPacket_, MyControl, and, using dc, MyChannel_.
   * 
   * @param dc - a reference to the related DataChannel.
   */
  public ControlChannel(DataChannel dc)
  {
    super();
    RTCPPacket_    = new RTCPPacket();
    MyControl_     = new Control(this);
    MyChannel_     = dc;

    dc.setControl(MyControl_);
  }

  /**
   * abstract method receive, which allows any transportlayer beneath.
   * Transportlayers could be UDP, TCP, ATM etc.
   */
  public abstract void receive();

  /**
   * abstract method send, which allows any transportlayer beneath.
   * Transportlayers could be UDP, TCP, ATM etc.
   *
   * @param packet - The RTCPPacket to be send.
   */
  public abstract void send(RTCPPacket packet);


  /**
   * the run-method issues receiver-reports on a regular basis.
   * The algorithm for Multicast-sessions, as seen in the specification, is not realized herin, because for now
   * we only have two-party unicast here.<br>
   * -The first RR-RTCP-Packet is send after a delay of 5 seconds. <br> 
   * -The following RR-Packets are send appropiate, as not to exceed the recommended 3.75% of the available bandwidth.<br>
   * -The BYE-Packet is send immediately.(This Packet will not be generated in this method)<br>
   * -Any APP-Packets are also send immediately. This should not become a bandwidth-problem, because all APP-Packets are
   * generated using the GUI of the Application. It would not make much sense, to generate Packets every second.<br>
   */
  public void run()
  {
    Random rand = new Random(System.currentTimeMillis());

     try { sleep(5000); }
     catch(InterruptedException e) { System.out.println("Ein Interrupt in ControlChannel: " +e.getMessage()); }

    while(Running_)
    {
      try
      {
         // [PENDING]: solange wir keine Infos zur Bandbreite haben werden RR-Packete alle 0-5 Sekunden generiert.
        sleep(rand.nextInt(5001));
      }
      catch(InterruptedException e)
      { System.out.println("Ein Interrupt in ControlChannel - while-Schleife: " +e.getMessage()); }

      MyChannel_.ExpectedInterval_ = MyControl_.HighestSeqNum_ - MyChannel_.ExpectedPrior_;
      MyChannel_.ReceivedInterval_ = MyChannel_.Received_      - MyChannel_.ReceivedPrior_;
      MyControl_.PacketsLost_      = MyControl_.HighestSeqNum_ - MyChannel_.Received_;

      if((MyControl_.HighestSeqNum_!=0) && (MyChannel_.ExpectedInterval_!=0))
      System.out.println("Overall fraction lost: " +((1f-((float)MyChannel_.Received_/(float)MyControl_.HighestSeqNum_))*100f) 
                        +"% Fraction lost in last RTCP-intervall: " 
                        +((1f-((float)MyChannel_.ReceivedInterval_/(float)MyChannel_.ExpectedInterval_))*100f)
                        +"%");
      MyControl_.sendReceiverReport(MyChannel_.Ssrc_, 
                                   (MyChannel_.ExpectedInterval_ - MyChannel_.ReceivedInterval_), 
                                    MyChannel_.ExpectedInterval_);

      MyChannel_.ExpectedPrior_ = MyControl_.HighestSeqNum_;
      MyChannel_.ReceivedPrior_ = MyChannel_.Received_;
    }
  }


  public void shutdown()
  {
    Running_ = false;
  }

}
