package bonn.infoIV.v3d2.rtp;

import bonn.infoIV.v3d2.buffer.SyncBuffer;
import bonn.infoIV.v3d2.buffer.SyncBuffer3D;
import bonn.infoIV.v3d2.buffer.Object3D;
import mpi_graphics.PM_Anim_Bonn;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.net.InetAddress;
import java.net.SocketException;
import java.util.Date;

public class DataChannel3D extends UDPDataChannel
{
  private static final int OVERFLOW = 0x10000; //

  private int TempSeqNum_      = 0; // The SeqNum of the actual RTPPacket 
  private int WrapCount_       = 0; // If TempSeqNum_ wraps, this counter is increased by 1
  private int LastValidSeqNum_ = 0; // Currently the RTP-Packets MUST come in the right order, to be processed!!!

  protected int FrameCounter_  = 0; // this counter counts the frames

  private boolean Frame_ = true; // Frame determines, into which array to put the incomming data TRUE==FRAME_1
  private boolean FirstPacket_ = true;      // The first packet has to be special treated


  public DataChannel3D(InetAddress dest, int destPort, int port, SyncBuffer3D buffer) throws SocketException
  {
    super(dest, destPort, port, buffer);
  }


  public void processPacket(DataInputStream in)
  {
    RTPPacket_.readPacket(in);
    Received_++;

    // computing the extended SequenceNumber, heeding the wraparount
    TempSeqNum_ = RTPPacket_.getHeader().getSequenceNumber();
    //System.out.println("SequNum: " +TempSeqNum_);
    if(TempSeqNum_==0) WrapCount_++;
    TempSeqNum_ +=WrapCount_*OVERFLOW;
    if(TempSeqNum_>MyControl_.HighestSeqNum_) MyControl_.HighestSeqNum_ = TempSeqNum_;

    // Now process the incoming payload, depending on the payload-type
    switch(RTPPacket_.getHeader().getPayloadType())
    {
      case PayloadType.VERTEX_SPLITS:
        if(LastValidSeqNum_==(TempSeqNum_-1))
	{
          ((SyncBuffer3D)MyBuffer_).putVertexSplits(RTPPacket_.getPayloadData(), Frame_, SyncBuffer.OBJECT);
          LastValidSeqNum_++;
        }
        else
	{
          LastValidSeqNum_ = 0;
        }
        break;

      case PayloadType.BASE_VERTICES:
        if(LastValidSeqNum_==(TempSeqNum_-1))
	{
          ((SyncBuffer3D)MyBuffer_).putBaseMeshVertices(RTPPacket_.getPayloadData(), Frame_, SyncBuffer.OBJECT);
          LastValidSeqNum_++;
	}
        else LastValidSeqNum_ = 0;
        break;

      case PayloadType.BASE_TRIANGLES:
        if(LastValidSeqNum_==(TempSeqNum_-1))
	{
          ((SyncBuffer3D)MyBuffer_).putBaseMeshTriangles(RTPPacket_.getPayloadData(), Frame_, SyncBuffer.OBJECT);
          LastValidSeqNum_++;
	}
        else LastValidSeqNum_ = 0;
        break;

      case PayloadType.PM_INFO:
        LastValidSeqNum_ = TempSeqNum_;
        FrameCounter_++;
        if(FirstPacket_) // First time, do not interrupt Consumer!!!
	{
          Ssrc_ = RTPPacket_.getHeader().getSSRC(); // You must set this only once, as we have only unicast 
          FirstPacket_ = false;
	}
        else // Now, each time you start a new Frame, check, which FRAME to write in
	{
          MyBuffer_.setStatus(Frame_, SyncBuffer.OBJECT); // set Status to readable
	  //System.out.println(new Date() + " Producer setzt Status auf Read bei Frame: " +Frame_);
          synchronized (this)  // as we have written a Frame, notify the Consumer
          {
            notify();
	    //System.out.println(new Date() + " Producer notifiziert fuer Frame: " +Frame_);
          }
          Frame_ = !Frame_;   // a new Frame --> a new Buffer; toggle to other buffer
	  //System.out.println(new Date() + " Producer wartet auf Frame: " +Frame_);
          if(!MyBuffer_.getStatus(Frame_, SyncBuffer.OBJECT)) ((PM_Anim_Bonn)MyConsumer_).waitForMe();
	  //System.out.println(new Date() + " Producer fertig...");
        }
        ((SyncBuffer3D)MyBuffer_).putInfo(RTPPacket_.getPayloadData(), Frame_, SyncBuffer.OBJECT);
        break;
      default:
        break;
    }
  }


  public synchronized void waitForMe()
  {
    try { wait(); } 
    catch (InterruptedException e) { }
  }

  public void createPacket(DataOutputStream out) {}
}


// processPacket
//    int offset = 0;
//     switch(RTPPacket_.getHeader().getPayloadType())
//     {
//       case PayloadType.VERTEX_SPLITS:
//         System.out.println("The Payloadtype is VERTEX_SPLITS");
//         MyBuffer_.putVertexSplits(RTPPacket_.getPayloadData(), 0, 0);
//         break;
//       case PayloadType.BASE_VERTICES:
//         System.out.println("The Payloadtype is BASE_VERTICES");  
//         MyBuffer_.putBaseMesh(RTPPacket_.getPayloadData(), 0, 0);
//         break;
//       case PayloadType.BASE_TRIANGLES:
//         System.out.println("The Payloadtype is BASE_TRIANGLES");
//         MyBuffer_.putBaseMesh(RTPPacket_.getPayloadData(), 0, 0);
//         break;
//       case PayloadType.PM_INFO:
//         System.out.println("The Payloadtype is PM_INFO");
//         MyBuffer_.putInfo(RTPPacket_.getPayloadData(), 0);
//         break;
//       default:
//         break;
//     }

//     if(RTPPacket_.getHeader().getRTPVersion()==RTPHeader.RTP_VERSION_2)
//       System.out.print("RTP_VERSION_2:");
//     System.out.print(":Padding:" +RTPPacket_.getHeader().getPadding());
//     System.out.print(":Extension:" +RTPPacket_.getHeader().getExtension());
//     System.out.print(":CSRC count:" +RTPPacket_.getHeader().getCSRCCount());
//     System.out.print(":Marker:" +RTPPacket_.getHeader().getMarker());
//     if(RTPPacket_.getHeader().getPayloadType()==PayloadType.VERTEX_SPLITS) 
//       System.out.print(":PT:VERTEX_SPLITS ");
//     else if(RTPPacket_.getHeader().getPayloadType()==PayloadType.PM_INFO) 
//       System.out.print(":PT:PM_INFO       ");
//     else if(RTPPacket_.getHeader().getPayloadType()==PayloadType.BASE_VERTICES) 
//       System.out.print(":PT:BASE_VERTICES ");
//     else if(RTPPacket_.getHeader().getPayloadType()==PayloadType.BASE_TRIANGLES) 
//       System.out.print(":PT:BASE_TRIANGLES");
//     System.out.print(":seqNum:" +RTPPacket_.getHeader().getSequenceNumber());
    //    System.out.println(":SSRC:" +RTPPacket_.getHeader().getSSRC());
    // [PENDING] Ueberpruefe ob es fuer uns ist, mittels Header
    // evtl ist es sogar besser erst nur den Header zu lesen und dann den Rest!!!
    // Der offset wird aus der SequenceNumber des Headers berechnet (Packetreihenfolge)
    //    if (FrameCounter_<RTPPacket_.getHeader().getTimeStamp()) System.out.println(":Frame no:" +FrameCounter_++);



