package bonn.infoIV.v3d2.rtp;

import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.SocketException;
import java.net.InetAddress;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;

public abstract class UDPControlChannel extends ControlChannel
{
  public final byte ChannelType_ = UDP_CHANNEL;

  protected DatagramSocket Socket_ = null;
  protected DatagramPacket PacketToSend_ = null;

  public UDPControlChannel(InetAddress destAddress, int destPort, int port, DataChannel dc) throws SocketException
  {
    super(dc);
    Socket_ = new DatagramSocket(port);
    Socket_.setReceiveBufferSize(RTCPPacket.MIN_PACKAGE_SIZE*BUFFER_SIZE);
    Socket_.setSendBufferSize(RTCPPacket.MIN_PACKAGE_SIZE*BUFFER_SIZE);
    Socket_.connect(destAddress, destPort);
    PacketToSend_ = new DatagramPacket(new byte[RTCPPacket.MIN_PACKAGE_SIZE*BUFFER_SIZE], 
                                       RTCPPacket.MIN_PACKAGE_SIZE*BUFFER_SIZE,
                                       destAddress,
                                       destPort);
  }


  public void receive()
  {
    DatagramPacket packet = null;
    try
    {
      packet = new DatagramPacket(new byte[Socket_.getReceiveBufferSize()], Socket_.getReceiveBufferSize());
      Socket_.receive(packet);
    }
    catch(Exception e)
    {
      System.err.println("[Error]: receiving RTCP-DatagramPacket: " +e.getMessage());
      System.exit(0);
    }
    processPacket(new DataInputStream(new ByteArrayInputStream(packet.getData())));
  }


  public void send(RTCPPacket payload)
  {
    byte data[] = payload.getAsByteArray();
    PacketToSend_.setData(data);

    // TD: Hier fehlte das Setzen der Lnge -> der Server bekam ein zu kurzes Paket
    //     und damit nach dem Header nur noch Mll!
    PacketToSend_.setLength(data.length);  

    try
    {
      Socket_.send(PacketToSend_);
    }
    catch(IOException e)
    {
      System.err.println("[Error]: sending RTCP-DatagramPacket: " +e.getMessage());
    }
  }

  public void shutdown() 
  {
    super.shutdown();
    Socket_.close();
  }

}
