package mpi_graphics;

import javax.media.j3d.ColoringAttributes;

public class Manifold_Mesh_Geometry extends Mesh_Geometry
{
    protected int[][] triangle_neighbor_array;
    protected int[] start_triangle_list;
    // C-Compatible
    protected int[] neighbour_triangle_counter;
    protected int[] neighbour_triangle_act_pos;
    protected int[] neighbour_triangle_cache;
    
    // Perfomance Cache
    protected byte [][] pos_cache;
    
    // Variablen fuer Methoden
    private int[] v_array = new int[3];
    private float[] point0 = new float[3];
    private float[] point1 = new float[3];
    private float[] point2 = new float[3];
    private float[] v1 = new float[3];
    private float[] v2 = new float[3];
    private float[] n_f= new float[3];
    private float[] n_f_afterburner;
    protected float[] n_f_afterburner_vertex;
 
// ************************************************************************************
// Manifold_Mesh_Geometry
// ************************************************************************************
    public Manifold_Mesh_Geometry(Mesh_Info info, Mesh_Info base_info, int shade_modus)
    {
	super(info,base_info,shade_modus);
	
	triangle_neighbor_array=new int[num_triangles_max][3];
	start_triangle_list=new int[num_vertices_max];
	// Hier reicht eigenlich die groesse des Maximalen Base-Mesh!!!
	neighbour_triangle_counter=new int[num_triangles_act];
	neighbour_triangle_act_pos=new int[num_triangles_act];
	neighbour_triangle_cache=new int[num_triangles_act*3];
	
	// Perfomance Cache
	pos_cache=new byte[num_triangles_max][3];
	
	n_f_afterburner = new float[3*num_triangles_max];
	
	if (shade_modus==ColoringAttributes.SHADE_GOURAUD)
	    n_f_afterburner_vertex = new float[3*num_vertices_max];
    }
        
// ************************************************************************************
// set_triangle_neighbor
// ************************************************************************************
    final public void set_triangle_neighbor(int t_number, int pos, int t_neighbor)
    {
	triangle_neighbor_array[t_number][pos]=t_neighbor;
    }
    
// ************************************************************************************
// get_triangle_neighbor
// ************************************************************************************
    final public int get_triangle_neighbor(int t_number, int pos)
    {
	return triangle_neighbor_array[t_number][pos];
    } 
    
// ************************************************************************************
// get_triangle_neighbors
// ************************************************************************************
    final public void get_triangle_neighbors(int t_number, int[] n_indices)
    {
	n_indices[0]=triangle_neighbor_array[t_number][0];
	n_indices[1]=triangle_neighbor_array[t_number][1];
	n_indices[2]=triangle_neighbor_array[t_number][2];
    } 
    
// ************************************************************************************
// build_neighbour_triangle_list
// ************************************************************************************
    public void build_neighbour_triangle_list()
    {
	int i;
	int sum;
	int last;
	
	// loesche letzten Eintrag
	for(i=0;i<num_triangles_act;i++)
	    {
		neighbour_triangle_counter[i]=0;
	    }

	// Ermittle Valenzen
	for(i=0;i<num_triangles_act;i++)
	    {
		get_vertex_indices(i,v_array);
		neighbour_triangle_counter[v_array[0]]++;
		neighbour_triangle_counter[v_array[1]]++;
		neighbour_triangle_counter[v_array[2]]++;
	    }

	sum=0;
	for(i=0;i<num_triangles_act;i++)
	    {
		last=neighbour_triangle_counter[i];
		neighbour_triangle_counter[i]=sum;
		neighbour_triangle_act_pos[i]=sum;
		sum+=last;
	    }
	    
	if (sum!=3*num_triangles_act)
	    Global.Error_Occured("Unmoeglicher Zustand!!!");

	for(i=0;i<num_triangles_act;i++)
	    {
		get_vertex_indices(i,v_array);

		neighbour_triangle_cache[neighbour_triangle_act_pos[v_array[0]]]=i;
		neighbour_triangle_act_pos[v_array[0]]++;
		neighbour_triangle_cache[neighbour_triangle_act_pos[v_array[1]]]=(i | POS_1_BIT);
		neighbour_triangle_act_pos[v_array[1]]++;
		neighbour_triangle_cache[neighbour_triangle_act_pos[v_array[2]]]=(i | POS_2_BIT);	
		neighbour_triangle_act_pos[v_array[2]]++;
	    }
	    
	for(i=0;i<num_vertices_act;i++)
	    {
		start_triangle_list[i]=neighbour_triangle_cache[neighbour_triangle_counter[i]];
	    }	    	 		 	
    }
  
// ************************************************************************************
// build_triangle_connections
// ************************************************************************************
    public void build_triangle_connections()
    {
	int i;

	//System.out.println("Erzeuge Dreiecks-Nachbarschaften");

	for (i=0;i<num_triangles_act;i++)
	    {
		get_vertex_indices(i,v_array);

		// Welche Dreiecke grenzen an?
		
		set_edge_neighbourhood(i,v_array[0],v_array[1],2);
		//System.out.println(inachbar);
		
		set_edge_neighbourhood(i,v_array[1],v_array[2],0);
		//System.out.println(inachbar);

		set_edge_neighbourhood(i,v_array[2],v_array[0],1);
		//System.out.println(inachbar);
	    }
    }
 
// ************************************************************************************
// set_edge_neighbourhood
// ************************************************************************************
    final public void set_edge_neighbourhood(int idreieck, int v_a_0, int v_a_1, int v_pos)
    {
	int i;
	int i_start,i_end;
	int act_dreieck;
	int act_pos;
	int v_next;
	
	i_start=neighbour_triangle_counter[v_a_0];
	i_end=neighbour_triangle_act_pos[v_a_0];
	
	for (i=i_start;i<i_end;i++)
	{
	    act_dreieck=neighbour_triangle_cache[i];
	    act_pos=act_dreieck>>>30;
	    act_dreieck=act_dreieck&POS_BLEND;
	    
	    v_next=get_vertex_index(act_dreieck,(act_pos+2)%3);
	    
	    if (v_next!=v_a_1)
		continue;
	    	
	    // Das Dreieck grenzt an die Kante an!
	    set_triangle_neighbor(idreieck, v_pos, act_dreieck);
	    pos_cache[idreieck][v_pos]=(byte)(act_pos);
	    return;
	}
	/*
	act_pos=(v_pos+1)%3;
	if (act_pos==POS_0)
	    start_triangle_list[v_a_0]=idreieck;
	else
	if (act_pos==POS_1)
	    start_triangle_list[v_a_0]=(idreieck | POS_1_BIT);
	else
	if (act_pos==POS_2)
	    start_triangle_list[v_a_0]=(idreieck | POS_2_BIT);
	else
	    Global.Error_Occured("Unmoeglicher Zustand!!!");
	*/
	set_triangle_neighbor(idreieck, v_pos, NO_TRIANGLE);
	return;
    }
    
// ************************************************************************************
// calculate_triangle_normal
// ************************************************************************************
    final public void calculate_triangle_normal(int t_number, float[] n_f)
    {
	float length;
	
	get_vertex_indices(t_number,v_array);
	get_vertex(v_array[0],point0);
	get_vertex(v_array[1],point1);
	get_vertex(v_array[2],point2);
	    
	v1[0]=point1[0]-point0[0];
	v1[1]=point1[1]-point0[1];
	v1[2]=point1[2]-point0[2];
	v2[0]=point2[0]-point0[0];
	v2[1]=point2[1]-point0[1];
	v2[2]=point2[2]-point0[2];
	
	n_f[0]=v1[1]*v2[2]-v1[2]*v2[1];
	n_f[1]=-(v1[0]*v2[2]-v1[2]*v2[0]);
	n_f[2]=v1[0]*v2[1]-v1[1]*v2[0];

	length=(float)Math.sqrt(n_f[0]*n_f[0]+n_f[1]*n_f[1]+n_f[2]*n_f[2]);
	
	n_f[0]=n_f[0]/length;
	n_f[1]=n_f[1]/length;
	n_f[2]=n_f[2]/length;
	//set_normal(t_number,n_f);	
    }
    
// ************************************************************************************
// calculate_triangle_normal_for_gourand
// ************************************************************************************
    final public void calculate_triangle_normal_for_gourand(int i)
    {
	get_vertex_indices(i,v_array);
	get_vertex(v_array[0],point0);
	get_vertex(v_array[1],point1);
	get_vertex(v_array[2],point2);
	    
	v1[0]=point1[0]-point0[0];
	v1[1]=point1[1]-point0[1];
	v1[2]=point1[2]-point0[2];
	v2[0]=point2[0]-point0[0];
	v2[1]=point2[1]-point0[1];
	v2[2]=point2[2]-point0[2];
	
	n_f_afterburner[3*i+0]=v1[1]*v2[2]-v1[2]*v2[1];
	n_f_afterburner[3*i+1]=-(v1[0]*v2[2]-v1[2]*v2[0]);
	n_f_afterburner[3*i+2]=v1[0]*v2[1]-v1[1]*v2[0];
    }
    
// ************************************************************************************
// calculate_normals
// ************************************************************************************
    final public void calculate_normals()
    {
	int i;
	
	if (shade_modus==ColoringAttributes.SHADE_GOURAUD)
	{
	    for (i=0;i<num_triangles_act;i++)
		calculate_triangle_normal_for_gourand(i);
	    
	    for (i=0;i<num_vertices_act;i++)
		transform_triangle_to_vertex_normal(i);

	    set_normals(n_f_afterburner_vertex);
	}
	else
	{
	    for (i=0;i<num_triangles_act;i++)
	    {
		calculate_triangle_normal(i,n_f);
	
		n_f_afterburner[3*i+0]=n_f[0];
		n_f_afterburner[3*i+1]=n_f[1];
		n_f_afterburner[3*i+2]=n_f[2];
	    }
	    
	    set_normals(n_f_afterburner);
	}
    }
    
// ************************************************************************************
// transform_triangle_to_vertex_normal
// ************************************************************************************
    public void transform_triangle_to_vertex_normal(int vertex)
    {
	int act_triangle_number;
	int old_triangle_number;
	int act_pos_1;
	int start_act_pos_1;
	int start_triangle_number;
	float length;
	
	    start_triangle_number=start_triangle_list[vertex];
	    start_act_pos_1=((start_triangle_number>>>30)+1)%3;
	    start_triangle_number=start_triangle_number&POS_BLEND;
	    act_triangle_number=start_triangle_number;
	    act_pos_1=start_act_pos_1;

	    n_f[0]=0.0f;
	    n_f[1]=0.0f;
	    n_f[2]=0.0f;
	    // Suche vorwaerts
	    do
	    {
		n_f[0]+=n_f_afterburner[3*act_triangle_number+0];
		n_f[1]+=n_f_afterburner[3*act_triangle_number+1];
		n_f[2]+=n_f_afterburner[3*act_triangle_number+2];

		old_triangle_number=act_triangle_number;
		act_triangle_number=triangle_neighbor_array[act_triangle_number][act_pos_1];
		act_pos_1=pos_cache[old_triangle_number][act_pos_1];
		
		if (act_triangle_number==NO_TRIANGLE);
		    break;
	    }
	    while (act_triangle_number!=start_triangle_number);
	    
	    if (act_triangle_number==NO_TRIANGLE)
	    {
		act_triangle_number=start_triangle_number;
		act_pos_1=start_act_pos_1;
	
		// Suche rueckwaerts
		do
		{
		    old_triangle_number=act_triangle_number;
		    act_triangle_number=triangle_neighbor_array[act_triangle_number][(act_pos_1+1)%3];
		    act_pos_1=(pos_cache[old_triangle_number][(act_pos_1+1)%3]+1)%3;

		    if (act_triangle_number==NO_TRIANGLE)
			break;

		    n_f[0]+=n_f_afterburner[3*act_triangle_number+0];
		    n_f[1]+=n_f_afterburner[3*act_triangle_number+1];
		    n_f[2]+=n_f_afterburner[3*act_triangle_number+2];
		}
		while (true);
	    }
	    
	    length=(float)Math.sqrt(n_f[0]*n_f[0]+n_f[1]*n_f[1]+n_f[2]*n_f[2]);
	    
	    n_f_afterburner_vertex[3*vertex+0]=n_f[0]/length;
	    n_f_afterburner_vertex[3*vertex+1]=n_f[1]/length;
	    n_f_afterburner_vertex[3*vertex+2]=n_f[2]/length;
    }

// ************************************************************************************
// calculate_edge_number
// ************************************************************************************
    final public int calculate_edge_number()
    {
	int i,act_pos;
	int[] neighbor_indices = new int[3];
	int edge_counter=0;
	
	for (i=0;i<num_triangles_act;i++)
	{
	    get_triangle_neighbors(i, neighbor_indices);
	    
	    for (act_pos=0;act_pos<3;act_pos++)
	    {	
		if ( (neighbor_indices[act_pos]==NO_TRIANGLE) || (neighbor_indices[act_pos]>i) )
		    edge_counter++;
	    }	    
	}
	
	return edge_counter;
    }    
// ************************************************************************************
// check_position_cache
// ************************************************************************************
    public void check_position_cache(int bound)
    {
	int i;
	int[] counter = new int[3];
	int dreieck_counter=0;
	int idreieck;
	int xyz;
	
	for (i=0;i<bound;i++)
	{
	    counter[pos_cache[i][0]]++;
	    counter[pos_cache[i][1]]++;
	    counter[pos_cache[i][2]]++;
	}
	
	System.out.println(counter[0]);
	System.out.println(counter[1]);
	System.out.println(counter[2]);
	System.out.println();

	for (i=0;i<bound;i++)
	{
	    for (xyz=0;xyz<3;xyz++)
	    {
		idreieck=triangle_neighbor_array[i][xyz];
		dreieck_counter+=idreieck;
		if (pos_cache[idreieck][(pos_cache[i][xyz]+1)%3]!=(xyz+2)%3)
		    Global.Error_Occured("Unmoeglicher Zustand!!!");
	    }
	}
	
	System.out.println(dreieck_counter);
	System.out.println();
    }
}
