// SSI.hh --- SSI Class Header File (RFC XXXX)

#ifndef SSI_H
#define SSI_H

#include <map.h>
#include <RTP/common.hh>

const unsigned short RTP_MAX_DROPOUT = 1000;
const unsigned short RTP_MAX_MISORDER = 100;
const unsigned short RTP_MIN_SEQUENTIAL = 2;

class SSI
{
private:
  u_int32 ssrc;                           // Data source identifier
  u_int32 lsr;                            // Middle 32 bits of last sender report
  u_int16 max_seq;                        // Highest seq. number seen
  u_int32 cycles;                         // Shifted count of seq. number cycles
  u_int32 base_seq;                       // Base seq. number
  u_int32 bad_seq;                        // Last 'bad' seq. number + 1
  u_int32 jitter;                         // Estimated jitter
  u_int32 transit;                        // Relative transmission time for prev. packet
  u_int32 probation;                      // Sequential packets till source is valid
  u_int32 received;                       // Packets received
  u_int32 received_prior;                 // Packets received at last interval
  u_int32 expected_prior;                 // Packets expected at last interval
public:
  SSI();
  void init(u_int16);
  void update(u_int16);
  u_int32 fractionLost();
  u_int32 packetsLost();
  u_int32 interJitter();
  u_int32 lastSeqNum();
  u_int32 SSRC();
  u_int32 LSR();
  u_int32 DLSR();
};

typedef map<unsigned int, SSI> SSI_Table;

#endif
