#include <session.hh>

                            // TD: Parameter qosManager
Prog4D_Session::Prog4D_Session(Coral::QoSManager* qosManager)
{
  // Synchronization device
  sem_init(&startup, 0, 0);

  // Set parameters
  modus = RTP_MODUS_LOOP;

  timestamp = 1;

  // TD: bw_wanted, bw_user und bw_cur initialisieren
  bw_user = bw_wanted = bw_cur = 8 * START_BPS;   //  1 MBit

  bw_min = 0;
  bw_max = 8 * MAXIMUM_BPS;  // 20 MBit   TD: hier stand 15M, Client hat Scrollbar bis 20M ?!

  threshold_lower = 0.02;  // 2 %
  threshold_upper = 0.04;  // 4 %

  filter_parameter = 0.3;

  rate = 1;   // 1 FPS

  frac = 100;  // 100 % of vertex splits

  bw_unt = bw_cur / rate;

  tu = 1.0 / rate;

  packets = 0;

  max_pminfos   = MAX_PMINFOS;
  max_vertices  = MAX_VERTICES;
  max_triangles = MAX_TRIANGLES;
  max_splits    = MAX_SPLITS;


  // TD: Attribute setzen
  TrafficClass          = 0x00;
  ExtensionTrafficClass = 0x00;
  QoSMgr                = qosManager;

  // TD: Flow anmelden unter ID 1234
  if(QoSMgr != NULL) {
     Coral::ExtendedTransportInfo transportInfo;
     getTransportInfo(&transportInfo);
     transportInfo.SrcDest[0].Destination.reset();
     transportInfo.SrcDest[0].Source.reset();
     transportInfo.SrcDest[0].FlowLabel    = 0;
     transportInfo.SrcDest[0].TrafficClass = TrafficClass;

     transportInfo.SrcDest[1].Destination.reset();
     transportInfo.SrcDest[1].Source.reset();
     transportInfo.SrcDest[1].FlowLabel    = 0;
     transportInfo.SrcDest[1].TrafficClass = ExtensionTrafficClass;

     StreamIdentifier = QoSMgr->getID();
     QoSMgr->newStream(StreamIdentifier,transportInfo,(Coral::QoSManagerInterface*)this);
  }

/*
  Coral::TransportInfo ti;
  getTransportInfo(&ti);
  cout << ti << endl;
*/
  // TD: Ende ----------------------
}




// ###### TD: Neue Methoden fr QoS-Management ##############################

/*
   Diese Funktion wird vom QoS-Manager aufgerufen, wenn eine neue Bandbreite
   festgelegt werden soll. Diese Funktion wird immer aufgerufen, wenn man
   QoSMgr->streamFlow() oder QoSMgr->updateStream() aufruft. Spter soll der
   QoS-Manager als Thread implementiert werden. Dann wird diese Funktion
   auch bei nderungen automatisch vom QoS-Manager-Thread aufgerufen. Deshalb
   mu hier eventuell fr Synchronisation mit den Prog4D-Threads gesorgt werden!!!
*/
void Prog4D_Session::adaptTransportInfo(card32 flowIdentifier, Coral::ExtendedTransportInfo ti)
{
   bw_cur = min(((unsigned long)ti.WantedBytesPerSecond[0].getValue() +
                 (unsigned long)ti.WantedBytesPerSecond[1].getValue()) * 8,bw_wanted);
   bw_unt = bw_cur / rate;
   if(ti.SrcDest[0].IsValid == true)
      TrafficClass = ti.SrcDest[0].TrafficClass;
   if(ti.SrcDest[1].IsValid == true)
      ExtensionTrafficClass = ti.SrcDest[1].TrafficClass;

   char str[32];
   sprintf((char*)&str,"$%02x / $%02x",TrafficClass,ExtensionTrafficClass);

   cout << "NEUE BANDBREITE: " << bw_cur << " (mgr="
        << ((unsigned long)(ti.WantedBytesPerSecond[0].getValue() + ti.WantedBytesPerSecond[1].getValue()) * 8)
        << ") Bit/s, TrafficClass " << str << endl;
}


/*
   Diese Funktion leitet die gewnschte Bandbreite an das QoS-Management
   weiter. Das QoS-Management fhrt dann adaptTransportInfo() oben aus.
   Dort wird dann die neue Bandbreite wirklich gesetzt.
*/
void Prog4D_Session::managementUpdate()
{
   if(QoSMgr != NULL) {
      Coral::ExtendedTransportInfo transportInfo;
      getTransportInfo(&transportInfo);
      transportInfo.SrcDest[0].Destination.reset();
      transportInfo.SrcDest[0].Source.reset();
      transportInfo.SrcDest[0].FlowLabel    = 0;
      transportInfo.SrcDest[0].TrafficClass = TrafficClass;

      transportInfo.SrcDest[1].Destination.reset();
      transportInfo.SrcDest[1].Source.reset();
      transportInfo.SrcDest[1].FlowLabel    = 0;
      transportInfo.SrcDest[1].TrafficClass = ExtensionTrafficClass;

      transportInfo.WantedBytesPerSecond[0].setValue((card64)(BASE_BPS * rate));
      transportInfo.WantedBytesPerSecond[1].setValue((card64)(bw_wanted / 8) - (BASE_BPS * rate));

      QoSMgr->updateStream(StreamIdentifier,transportInfo);
         // QoSMgr ruft adaptTransportInfo() mit zugewiesener Bandbreite auf.
         // Erst dort wird die neue Bandbreite wirklich gesetzt.
   }
   else {                             // TD: Kein QoS-Manager -> direkt setzen
      bw_cur = bw_wanted;
      bw_unt = bw_cur / rate;

/*
      // TrafficClass-Test
      TrafficClass          = 36;
      ExtensionTrafficClass = 38;
*/

   }
}


/*
   Diese Funktion gibt QoS-Konstanten fr das QoS-Management zurck.
*/
void Prog4D_Session::getTransportInfoConstants(Coral::TransportInfo* transportInfo)
{
   // ====== Initialize wanted values ====================================
   transportInfo->WantedBytesPerSecond[0].setLimits(BASE_BPS,30 * BASE_BPS);
   transportInfo->WantedPacketsPerSecond[0].init((card32)-1,(card32)-1,(card32)-1);
   transportInfo->WantedMaxLossRate[0] = 0;

   transportInfo->WantedBytesPerSecond[1].setLimits(0,MAXIMUM_BPS - BASE_BPS);
   transportInfo->WantedPacketsPerSecond[1].init((card32)-1,(card32)-1,(card32)-1);
   transportInfo->WantedMaxLossRate[1] = 0;

   for(int i = 2;i < Coral::TransportInfoLevel::MaxQualityLayers;i++) {
      transportInfo->WantedBytesPerSecond[i].init(0,0,0);
      transportInfo->WantedPacketsPerSecond[i].init(0,0,0);
      transportInfo->WantedMaxLossRate[i] = 0;
   }

   transportInfo->WantedFramesPerSecond.setLimits(1,30);
   transportInfo->WantedMaxTransferDelay  = 1000 * 16;
   transportInfo->StartFramesPerSecond    = 1;

   transportInfo->Flags         = Coral::TransportInfo::TIF_None;
   transportInfo->QualityLevels = 1;
   transportInfo->QualityLayers = 2;
      

   // ====== Initialize level and layer =====================================
   transportInfo->QualityLevel[0].BytesPerSecondScale   = 65536;
   transportInfo->QualityLevel[0].PacketsPerSecondScale = 65536;
   transportInfo->QualityLevel[0].FramesPerSecondScale  = 65536;
   transportInfo->QualityLevel[0].FramesPerSecond       = 1;
   transportInfo->QualityLevel[0].MaxTransferDelay      = 16 * 1000;
   transportInfo->QualityLevel[0].Quality               = 255;
   transportInfo->QualityLevel[0].LevelUp               = 255;
   transportInfo->QualityLevel[0].LevelDown             = 255;
   transportInfo->QualityLevel[0].QualityLayers         = 2;

   transportInfo->QualityLevel[0].QualityLayer[0].BytesPerSecond   = BASE_BPS;
   transportInfo->QualityLevel[0].QualityLayer[0].PacketsPerSecond = (card32)-1;
   transportInfo->QualityLevel[0].QualityLayer[0].FrameSize        = BASE_BPS;
   transportInfo->QualityLevel[0].QualityLayer[0].MaxLossRate      = 0;
   transportInfo->QualityLevel[0].QualityLayer[0].Flags            = Coral::TransportInfoLayer::TILF_Static;

   transportInfo->QualityLevel[0].QualityLayer[1].BytesPerSecond   = RTP_MAX_PAYLOAD * 10;
   transportInfo->QualityLevel[0].QualityLayer[1].PacketsPerSecond = (card32)-1;
   transportInfo->QualityLevel[0].QualityLayer[1].FrameSize        = RTP_MAX_PAYLOAD * 10;
   transportInfo->QualityLevel[0].QualityLayer[1].MaxLossRate      = 25;
   transportInfo->QualityLevel[0].QualityLayer[1].Flags            = Coral::TransportInfoLayer::TILF_Dynamic;


   // ====== Initialize current settings ====================================
   transportInfo->CurrentSetting.BytesPerSecondScale   = 65536;
   transportInfo->CurrentSetting.PacketsPerSecondScale = 65536;
   transportInfo->CurrentSetting.FramesPerSecondScale  = 65536;
   transportInfo->CurrentSetting.FramesPerSecond       = rate;
   transportInfo->CurrentSetting.MaxTransferDelay      = 16 * 1000;
   transportInfo->CurrentSetting.Quality               = 255;
   transportInfo->CurrentSetting.LevelUp               = 255;
   transportInfo->CurrentSetting.LevelDown             = 255;
   transportInfo->CurrentSetting.QualityLayers         = 2;

   transportInfo->CurrentSetting.QualityLayer[0].FrameSize   = BASE_BPS;
   transportInfo->CurrentSetting.QualityLayer[0].MaxLossRate = 0;
   transportInfo->CurrentSetting.QualityLayer[0].Flags       = Coral::TransportInfoLayer::TILF_Static;

   transportInfo->CurrentSetting.QualityLayer[1].FrameSize   = EXT_BPS;
   transportInfo->CurrentSetting.QualityLayer[1].MaxLossRate = 25;
   transportInfo->CurrentSetting.QualityLayer[1].Flags       = Coral::TransportInfoLayer::TILF_Dynamic;

   transportInfo->CurrentSetting.QualityLayer[0].BytesPerSecond   = (card64)(BASE_BPS * rate);
   transportInfo->CurrentSetting.QualityLayer[1].BytesPerSecond   = (card64)(getBPS() / 8) - (BASE_BPS * rate);
   transportInfo->CurrentSetting.QualityLayer[0].PacketsPerSecond = (card32)-1;
   transportInfo->CurrentSetting.QualityLayer[1].PacketsPerSecond = (card32)-1;
}


/*
   Diese Funktion gibt QoS-Konstanten und Variablen fr das QoS-Management zurck.
*/
void Prog4D_Session::getTransportInfo(Coral::TransportInfo* transportInfo)
{
   getTransportInfoConstants(transportInfo);
   transportInfo->WantedBytesPerSecond[0].setValue((card64)(BASE_BPS * rate));
   transportInfo->WantedBytesPerSecond[1].setValue((card64)(getBPS() / 8) - (BASE_BPS * rate));
   transportInfo->WantedFramesPerSecond.setValue(rate);
}

// ###### TD: Ende ##########################################################




void Prog4D_Session::sendLoopRTP()
{
  // Prepare for transmission
  open("anim_compact.pm", ios::in | ios::binary);
  if (!*this)
    {
      cerr << "Could not open anim_compact.pm file !\n" << endl;
      exit(EXIT_FAILURE);
    };

  init();

  sem_wait(&startup);

  unsigned int   num;
  unsigned int   count;
  unsigned short bytes;

  Progressive_Packet ProgressivePacket;

  // Init time for first unit
  tn.update();

  // TD: Noch keine Traffic class
  card8 trafficClass = 0x00;
  card32 totalBPS    = 0;
  card32 totalFrames = 0;

  while(1)
    {
      /*
       * Order of execution is critical !
       */
      if (packets == 0)
	{
	  if(totalFrames >= rate) {
	     cout << totalBPS << " Bytes (" << (totalBPS * 8) << " Bits) in "
	          << totalFrames << " frames. Bandwidth: " << (bw_cur / 8)
	          << " Bytes (" << bw_cur << " Bits)." << endl;
             totalBPS    = 0;
             totalFrames = 0;
          }

          totalFrames++;
	  next();

          // Auf zu geringe Bandbreite berprfen
          if((bw_bm > bw_unt) && (bw_bm > bw_unt)) {
             cout << "Bandwidth too low -> pause." << endl;
             usleep(1000000);
             packets = 0;
             goto nextFrame;
          }

	  pminfo_count = 0;
	  vertex_count = 0;
	  triangle_count = 0;
	  vertex_split_count = 0;
	  timestamp = timestamp + 1;
	};
      if (pminfos != 0)
	{
	  ptype = RTP_PM_INFOS;

	  if (pminfos < max_pminfos) num = pminfos;
	  else num = max_pminfos;
	  pminfos = pminfos - num;
	  pminfo_count = pminfo_count + num;
	  bytes = num * sizeof(PMInfo);
	  count = pminfo_count;
	  
	  // TD:
	  trafficClass = TrafficClass; // Basisdaten
	}
      else if (vertices != 0)
	{
	  ptype = RTP_BASE_VERTICES;

	  if (vertices < max_vertices) num = vertices;
	  else num = max_vertices;
	  vertices = vertices - num;
	  vertex_count = vertex_count + num;
	  bytes = num * sizeof(Vertex);
	  count = vertex_count;

          // TD:
	  trafficClass = TrafficClass; // Basisdaten
	}
      else if (triangles != 0)
	{
	  ptype = RTP_BASE_TRIANGLES;

	  if (triangles < max_triangles) num = triangles;
	  else num = max_triangles;
	  triangles = triangles - num;
	  triangle_count = triangle_count + num;
	  bytes = num * sizeof(Triangle);
	  count = triangle_count;

         
          // TD:
	  trafficClass = TrafficClass; // Basisdaten
	}
      else if (vertex_splits != 0)
	{
	  ptype = RTP_VERTEX_SPLITS;

	  if (vertex_splits < max_splits) num = vertex_splits;
	  else num = max_splits;
	  vertex_splits = vertex_splits - num;
	  vertex_split_count = vertex_split_count + num;
	  bytes = num * sizeof(Vertex_Split);
	  count = vertex_split_count;

          // TD:
	  trafficClass = ExtensionTrafficClass; // Erweiterungsdaten
	}
      else
	{
	  cout << "Packetizer error !" << endl;
	  continue;
	}

      ProgressivePacket.setPayloadType(ptype);
      ProgressivePacket.setTimestamp(timestamp);
      ProgressivePacket.setNumber(count);
      ProgressivePacket.setLength(sizeof(RTP_Packet) + sizeof(unsigned int) + bytes);

      read(((char*) (&ProgressivePacket)) + sizeof(RTP_Packet) + sizeof(unsigned int), bytes);

      /*
       * First: We check if the current time is the time to
       * send the next unit. If NOT: We sleep until time has come.
       */
      if (tc.update() < tn)
	{
	  ts = tn - tc;
	  select(0, NULL, NULL, NULL, &ts);
	};

      /*
       * Third: We send the current packet until one of the
       * following constrains is reached:
       * a) The frame has been sent completly.
       * b) The bandwidth limit for this unit is reached.
       * c) The time for this unit is up.
       */

      // TD: TrafficClass an send() bergeben
      totalBPS += ProgressivePacket.getLength() + 20 + 8;  // addiere IP-Lnge + UDP-Lnge!!!
      if(send(&ProgressivePacket, ProgressivePacket.getLength(), trafficClass) != ProgressivePacket.getLength()) {
         cerr << "Client beendet? -> Warte 1 Sekunde..." << endl;
         usleep(1000000);
      }

      packets = packets - 1;

nextFrame:
      /*
       * Second: We calculate the absolute time on which the
       * following unit has to be sent. This is also the time
       * limit for the current unit.
       */
      tn = tn + ti;
    }
}

void Prog4D_Session::recvLoopRTP()
{
  // Handshake Packet
  RTCP_App RTCPApp;

  // Values for Handshake
  RTCPApp.val[3] = reverse((unsigned int) 5000000);
  RTCPApp.val[1] = reverse((unsigned int) 20);
  RTCPApp.val[2] = reverse((unsigned int) 100);
  RTCPApp.val[0] = reverse((unsigned int) true);

  // Send Handshake Packet
  send(&RTCPApp,0);

  Progressive_Packet ProgressivePacket;
  /*
  ofstream out;
  out.open("out.pm", ios::out | ios::binary);
  if (!out)
    {
      cerr << "Could not open file !\n" << endl;
      exit(EXIT_FAILURE);
    }
  else
    {
      char* file_sig = "#PMESH-C";
      unsigned short file_sig_len = reverse((unsigned short) 8);
      unsigned int file_frames = reverse((unsigned int) 247);

      out.write(&file_sig_len, sizeof(file_sig_len));
      out.write(file_sig, reverse(file_sig_len));
      out.write(&file_frames, sizeof(file_frames));
    }

  char* frame_sig = "#PMESH-B";
  unsigned short frame_sig_len = reverse((unsigned short) 8);
  */
  while(1)
    {
      unsigned short bytes = recv(&ProgressivePacket, sizeof(ProgressivePacket) - sizeof(unsigned short));

      switch(ProgressivePacket.getPayloadType())
	{
	case RTP_PM_INFOS       : //out.write(&frame_sig_len, sizeof(frame_sig_len));
	                          //out.write(frame_sig, reverse(frame_sig_len));
	                          //for(unsigned int i = 0; i < (bytes - sizeof(RTP_Packet) - 4) / sizeof(PMInfo); i++)
	                          //  out << ProgressivePacket.pminfo[i] << flush;
	                          break;

	case RTP_BASE_VERTICES  : //for(unsigned int i = 0; i < (bytes - sizeof(RTP_Packet) - 4) / sizeof(Vertex); i++)
	                          //  out << ProgressivePacket.vertex[i] << flush;
	                          break;

	case RTP_BASE_TRIANGLES : //for(unsigned int i = 0; i < (bytes - sizeof(RTP_Packet) - 4) / sizeof(Triangle); i++)
	                          //  out << ProgressivePacket.triangle[i] << flush;
	                          break;

	case RTP_VERTEX_SPLITS  : //for(unsigned int i = 0; i < (bytes - sizeof(RTP_Packet) - 4) / sizeof(Vertex_Split); i++)
	                          //  out << ProgressivePacket.split[i] << flush;
	                          break;

	default                 : cout << "Unknown Payload Type: Packet ignored !" << endl;
	                          break;
				  }
    }
}

void Prog4D_Session::feedback(RTCP_Packet* RTCPPacket)
{
  static int i=1;
  // cout << "Receiving RTCP Packet...[" << i++ << "]" << "-->";

  RTCP_Receiver_Report* RTCPReceiverReport;
  RTCP_App* RTCPApp;

  switch(RTCPPacket->getPacketType())
    {
    case RTCP_RR  : RTCPReceiverReport = (RTCP_Receiver_Report*) RTCPPacket;
                    adapt(RTCPReceiverReport->rr[0].getFractionLost());
                    break;

    case RTCP_APP : RTCPApp = (RTCP_App*) RTCPPacket;
                    switch(RTCPApp->getCount())
                    {
		    case RTCP_APP_HANDSHAKE : cout << endl;
		                              // TD: bw_user setzen
                                              bw_user = (unsigned long)reverse(RTCPApp->val[3]);
		                              setBPS((unsigned long) reverse(RTCPApp->val[3]));
		                              setRate((unsigned short) reverse(RTCPApp->val[1]));
			                      setFraction((unsigned short) reverse(RTCPApp->val[2]));
					      setModus((RTP_PlaybackModus) reverse(RTCPApp->val[0]));
					      sem_post(&startup);
					      break;

      		    case RTCP_APP_BPS       : // TD: bw_user setzen
      		                              bw_user = (unsigned long)reverse(RTCPApp->val[0]);
      		                              setBPS((unsigned long) reverse(RTCPApp->val[0]));
		                              break;

		    case RTCP_APP_RPS       : setRate((unsigned short) reverse(RTCPApp->val[0]));
		                              break;

		    case RTCP_APP_VSF       : setFraction((unsigned short) reverse(RTCPApp->val[0]));
		                              break;

		    case RTCP_APP_MODUS     : setModus((RTP_PlaybackModus) reverse(RTCPApp->val[0]));
		                              break;

		    default                 : cerr << "Unknown Appl. defined RTCP Packet Type: Packet ignored !" << endl;
		                              break;
		    }
		    break;
    default       : cerr << "Unhandeled RTCP Packet Type: Packet ignored !" << endl;
                    break;
    }


   // TD: Falls Server wieder neu gestartet wurde.
   sem_post(&startup);
}

void Prog4D_Session::adapt(float fraction)
{
  // TD: Mit QoS-Manager: Adapt-Funktionalitt von QoS-Manager bernommen!
  if(QoSMgr != NULL) {
     QoSMgr->setFractionLost(1234,fraction,1);
  }   

  // TD: Ohne QoS-Manager: wie bisher.
  else {

// Bandbreite-Rckschaltung *ohne* QoS-Manager mu noch berarbeitet werden,
// damit herabgesetzter Wert vom Client nicht sofort wieder auf ursprnglichen
// Wert gesetzt wird.

/*
     // Network state estimation based on loss rate
     NETWORK_STATE net_state;

     if (fraction > threshold_upper)
       net_state = CONGESTED;
     else if (fraction > threshold_lower)
       net_state = LOADED;
     else
       net_state = UNLOADED;

     // Making a bandwidth adjustment decision
     BANDWIDTH_ADJ bw_adj;

     if (net_state == CONGESTED)
       bw_adj = DECREASE;
     else if (net_state == LOADED)
       bw_adj = HOLD;
     else if (net_state == UNLOADED)
       bw_adj = INCREASE;

     unsigned long newBandwidth = getBPS(); // TD: Neue Variable

     // Calculation of allowed bandwidth.
     if (bw_adj == DECREASE)
       {
         //newBandwidth = max(getBPS() * 0.5, bw_min);  // TD: newBandwidth statt bw_cur
         if (getBPS() * 0.5 > bw_min)
        	newBandwidth = (unsigned long) (getBPS() * 0.75);
         else
        	newBandwidth = bw_min;
         //cout << "Decreasing...";
       }
     else if (bw_adj == INCREASE)
       {
         //newBandwidth = min(getBPS() + 5000, bw_max);  // TD: newBandwidth statt bw_cur
         if (getBPS() + 5000 < bw_max)
        	newBandwidth = getBPS() + 500000;
         else
        	newBandwidth = bw_max;
         //cout << "Increasing...";
       }
     else
       {
         //cout << "Holding..." << endl;
         return;
       }

     //printf("adapt: newBandwidth=%d  bw_user=%d\n",newBandwidth,bw_user);

     newBandwidth = min(bw_user,newBandwidth); // TD: Nicht hher als bw_user gehen!
     setBPS(newBandwidth);                     // TD: newBandwidth statt bw_cur
*/
  }
}

RTP_PlaybackModus Prog4D_Session::getModus()
{
  return modus;
}

void Prog4D_Session::setModus(RTP_PlaybackModus mod)
{
  modus = mod;

  if (modus == RTP_MODUS_SIMPLE)
    cout << "[" << "Loop Modus inactive" << "]" << endl;
  else if (modus == RTP_MODUS_LOOP)
    cout << "[" << "Loop Modus active" << "]" << endl;
}

unsigned long Prog4D_Session::getBPS() { return bw_cur; }

void Prog4D_Session::setBPS(unsigned long bits_per_second)
{
  // Set new value for bps
  if (bits_per_second < bw_min) bw_wanted = bw_min;
  else if (bits_per_second > bw_max) bw_wanted = bw_max;
  else bw_wanted = bits_per_second;

/*
  TD: Das darf jetzt erst in adaptTransportInfo() gesetzt werden!!!

  // Calculate new value for bw_unt
  bw_unt = bw_wanted / rate;
*/

  // TD: Hier haben wir jetzt nur *gewnschte* Bandbreite
  // cout << " Neue gewnschte Bandbreite: [" << bw_wanted << " Bit/s" << "]" << endl;

  // TD: Erst QoS-Manager fragen, ob neue Bandbreite berhaupt mglich,
  //     dann Bandbreite wirklich setzen.
  managementUpdate();  
}


unsigned short Prog4D_Session::getRate() { return rate; }

void Prog4D_Session::setRate(unsigned short rate_per_second)
{
  // Set new value for fps
  if (rate_per_second <= 0) rate = 1;
  else if (rate_per_second >= 30) rate = 30;         // TD: oberes Limit auf 30 statt 50.
  else rate = rate_per_second;

  // Calculate new value for bw_unt
  bw_unt = bw_cur / rate;

  // Calculate the transmission interval
  tu = 1.0 / rate;

  // cout << "[" << rate << " Frames/s" << "]" << endl;
}

unsigned short Prog4D_Session::getFraction()
{
  return frac;
}

void Prog4D_Session::setFraction(unsigned short fraction)
{
  // Set new value for fraction
  if (fraction <= 0) frac = 0;  // < statt <=, aber sonst unsinnig
  else if (fraction > 100) frac = 100;
  else frac = fraction;

  cout << "[" << frac << " % Vertex Splits/Frame" << "]" << endl;
}

void Prog4D_Session::init()
{
  /*
   * File starts with 14 bytes header. First two bytes are
   * the length of the following string and the next 8 bytes
   * are the string itself. Last 4 bytes of the 14 bytes file
   * header is the number of frames following.
   */
  unsigned short sig_len;

  read(&sig_len, sizeof(sig_len));

  sig_len = reverse(sig_len);

  char* file_signature = new char[sig_len + 1];

  read(file_signature, sig_len);

  file_signature[sig_len] = '\0';

  int ret = strcmp(FILE_SIGNATURE, file_signature);
  if (ret != 0)
    {
      cerr << "Invalid file header !\n" << endl;
      exit(EXIT_FAILURE);
    };
  delete[] file_signature;

  read(&frames, sizeof(frames));

  frames = reverse(frames);

  position = tellg();
}

void Prog4D_Session::next()
{
  // Check for end-of-file / end-of-sequence
  if (frames == 0)
    {
      if (modus == RTP_MODUS_LOOP)
	{
	  seekg(0); // rewind(); ssekg(ios::beg)
	  init();
	}
      else if (modus == RTP_MODUS_SIMPLE)
	{
	  close();
	  cout << "\n\nTransmission completed !\n\n" << endl;
	  exit(EXIT_SUCCESS);
	}
    };

  // Position file pointer to beginning of next frame
  seekg(position);

  // Update frame counter
  frames = frames - 1;

  /*
   * Frame starts with 22 bytes header. First two bytes are
   * the length of the following string and the next 8 bytes
   * are the string itself. After the string follows another
   * 12 bytes where the first 4 bytes are the number of vertices
   * the middle 4 bytes are the number of triangles and the
   * last 4 bytes are number of vertex splits.
   */
  unsigned short sig_len;

  read(&sig_len, sizeof(sig_len));

  sig_len = reverse(sig_len);

  char* frame_signature = new char[sig_len + 1];

  read(frame_signature, sig_len);

  frame_signature[sig_len] = '\0';

  int ret = strcmp(FRAME_SIGNATURE, frame_signature);
  if (ret != 0)
    {
      cerr << "Invalid frame header !\n" << endl;
      exit(EXIT_FAILURE);
    };
  delete[] frame_signature;

  // Remember actual file position
  streampos pos = tellg();

  read(&vertices,      sizeof(vertices));
  read(&triangles,     sizeof(triangles));
  read(&vertex_splits, sizeof(vertex_splits));

  // Reset the file pointer to remembered position
  seekg(pos);

  // Convert to host byte order
  pminfos       = 1;
  vertices      = reverse(vertices);
  triangles     = reverse(triangles);
  vertex_splits = reverse(vertex_splits);

//cout << vertices << " " << triangles << " " << vertex_splits << endl;

  // Calculate position of next frame
  position = pos                              +
             1             * sizeof(PMInfo)   +
             vertices      * sizeof(Vertex)   +
             triangles     * sizeof(Triangle) +
             vertex_splits * sizeof(Vertex_Split);

  // Calculate transmission interval
  calcTransInterval();
}

void Prog4D_Session::calcTransInterval()
{
  // Calculate fraction of vertex_splits to be transmitted
  vertex_splits = (unsigned int) (vertex_splits * frac / 100.0);

  // Calculate the bandwidth needed for the base mesh
  bw_bm = pminfos   * sizeof(PMInfo)   * 8 +
          vertices  * sizeof(Vertex)   * 8 +
          triangles * sizeof(Triangle) * 8;

  // Calculate the bandwidth of this mesh
  bw_ms = bw_bm + vertex_splits * sizeof(Vertex_Split) * 8;

  // Maybe the mesh is too big
  if (bw_ms > bw_unt)
    {
      // Maybe the base mesh is too big
      if (bw_bm > bw_unt)
	{
          // TD: auskommentiert, wird in Sendeschleife abgefangen
	  // cerr << "The base mesh is too big [skipping frame] !" << endl;

          // TD: Das folgende next() wrde eine Endlos-Rekursion auslsen,
          //     die mit Segfault endet, wenn der Stack berluft:
          //     next() -> calcTransInterval() -> next().
          // next();
	}
      else
	{
	  bw_vs = bw_unt - bw_bm;
	  vertex_splits = (unsigned int) (bw_vs / (sizeof(Vertex_Split) * 8));
	}
    }


  // ====== TD: Einige Konstanten berechnen =================================
  const unsigned int ext  = vertex_splits * sizeof(Vertex_Split);
  const unsigned int base = (bw_bm / 8);

  const unsigned int extPackets    = (unsigned int)ceil((double)(vertex_splits * sizeof(Vertex_Split)) / (RTP_MAX_PAYLOAD - NUMBER_SIZE));
  const unsigned int pmPackets     = (unsigned int)ceil((double)(pminfos * sizeof(PMInfo)) / (RTP_MAX_PAYLOAD - NUMBER_SIZE));
  const unsigned int vertexPackets = (unsigned int)ceil((double)(vertices * sizeof(Vertex)) / (RTP_MAX_PAYLOAD - NUMBER_SIZE));
  const unsigned int trianglePackets = (unsigned int)ceil((double)(triangles* sizeof(Triangle)) / (RTP_MAX_PAYLOAD - NUMBER_SIZE));
  const unsigned int basePackets   = pmPackets + vertexPackets + trianglePackets;

  // printf("Base=%d Ext=%d   PBase=%d PExt=%d  %d %d %d\n",base,ext,basePackets,extPackets,pmPackets,vertexPackets,trianglePackets);

  // ====== Auf Bandbreite-berschreitung testen, ggf. Vertex-Daten krzen ==
  int cut = ((int)(base + ext) + (basePackets + extPackets) * (20 + 8 + 12 + 4)) - (int)(bw_unt / 8);
  if(cut > 0) {
     const unsigned int vertexCut = (unsigned int)ceil((double)cut / (double)sizeof(Vertex_Split));
     // printf("VertexCut=%d\n",vertexCut);

     vertex_splits -= vertexCut;
     bw_vs -= vertexCut * 8;
  }

  // ========================================================================

  // Calculate number of packets
  packets = (unsigned int) (pminfos       / max_pminfos)   +
            (unsigned int) (vertices      / max_vertices)  +
            (unsigned int) (triangles     / max_triangles) +
            (unsigned int) (vertex_splits / max_splits);

  if ((pminfos       % max_pminfos)   != 0) packets = packets + 1;
  if ((vertices      % max_vertices)  != 0) packets = packets + 1;
  if ((triangles     % max_triangles) != 0) packets = packets + 1;
  if ((vertex_splits % max_splits)    != 0) packets = packets + 1;


/*
printf("%d <-> %d\n",extPackets+basePackets,packets);
if(extPackets+basePackets < packets) {
   puts("PACKET-BERECHNUNG!!!");
}
*/


  ti = tu / packets;
}

unsigned short Progressive_Packet::getLength() { return length; }

void Progressive_Packet::setLength(unsigned short len) { length = len; }

u_int32 Progressive_Packet::getNumber() { return number; }

void Progressive_Packet::setNumber(u_int32 num) { number = reverse(num); }

istream& operator>>(istream& input, PMInfo& info)
{
  input >> info.base_vertices;
  input >> info.base_triangles;
  input >> info.vertex_splits;

  return input;
}

istream& operator>>(istream& input, Vertex& vertex)
{
  input >> vertex.x;
  input >> vertex.y;
  input >> vertex.z;

  return input;
}

istream& operator>>(istream& input, Triangle& triangle)
{
  input >> triangle.a;
  input >> triangle.b;
  input >> triangle.c;

  return input;
}

istream& operator>>(istream& input, Vertex_Split& vertex_split)
{
  input >> vertex_split.a;
  input >> vertex_split.b;
  input >> vertex_split.c;
  input >> vertex_split.v;

  return input;
}

ostream& operator<<(ostream& output, PMInfo& info)
{
  output << reverse(info.base_vertices)  << " ";
  output << reverse(info.base_triangles) << " ";
  output << reverse(info.vertex_splits);

  return output;
}

ostream& operator<<(ostream& output, Vertex& vertex)
{
  output << reverse(vertex.x) << " ";
  output << reverse(vertex.y) << " ";
  output << reverse(vertex.z);

  return output;
}

ostream& operator<<(ostream& output, Triangle& triangle)
{
  output << reverse(triangle.a) << " ";
  output << reverse(triangle.b) << " ";
  output << reverse(triangle.c);

  return output;
}

ostream& operator<<(ostream& output, Vertex_Split& vertex_split)
{
  output << reverse(vertex_split.a) << " ";
  output << reverse(vertex_split.b) << " ";
  output << reverse(vertex_split.c) << " ";
  output << vertex_split.v;

  return output;
}
/*
ostream& operator<<(ostream& output, PMInfo& info)
{
  output.write(&info.base_vertices,  sizeof(info.base_vertices));
  output.write(&info.base_triangles, sizeof(info.base_triangles));
  output.write(&info.vertex_splits,  sizeof(info.vertex_splits));

  return output;
}

ostream& operator<<(ostream& output, Vertex& vertex)
{
  output.write(&vertex.x, sizeof(vertex.x));
  output.write(&vertex.y, sizeof(vertex.y));
  output.write(&vertex.z, sizeof(vertex.z));

  return output;
}

ostream& operator<<(ostream& output, Triangle& triangle)
{
  output.write(&triangle.a, sizeof(triangle.a));
  output.write(&triangle.b, sizeof(triangle.b));
  output.write(&triangle.c, sizeof(triangle.c));

  return output;
}

ostream& operator<<(ostream& output, Vertex_Split& vertex_split)
{
  output.write(&vertex_split.a, sizeof(vertex_split.a));
  output.write(&vertex_split.b, sizeof(vertex_split.b));
  output.write(&vertex_split.c, sizeof(vertex_split.c));
  output << vertex_split.v;

  return output;
}
*/
