#include <socket/socket.hh>


#include <netinet/in.h>


Socket::Socket()
{
  sd = -1;
  domain = -1;
  type = -1;
  protocol = -1;
  backlog = -1;
  send_flags = 0;
  recv_flags = 0;
}

int Socket::create(int s_domain, int s_type, int s_protocol)
{
  int ret = socket(s_domain, s_type, s_protocol);

  // ret == -1 on failure
  if (ret == -1)
    perror("socket()");
  // ret != -1 on success
  else
    {
      sd = ret;
      domain = s_domain;
      type = s_type;
      protocol = s_protocol;
    }

  return ret;
}

int Socket::assignName(struct sockaddr *addr, int addrlen)
{
  int ret = bind(sd, addr, addrlen);

  // ret == -1 on failure
  if (ret == -1)
    perror("bind()");
  // ret == 0 on success

  return ret;
}

int Socket::makeListening(int s_backlog)
{
  int ret = listen(sd, s_backlog);

  // ret == -1 on failure
  if (ret == -1)
    perror("listen()");
  // ret == 0 on success
  else
    backlog = s_backlog;

  return ret;
}

int Socket::acceptConnection(struct sockaddr *addr, unsigned int *addrlen)
{
  int ret = accept(sd, addr, addrlen);

  // ret == -1 on failure
  if (ret == -1)
    perror("accept()");
  // ret >= 0 on success
  else
    sd = ret;

  return ret;
}

int Socket::initConnection(struct sockaddr *addr, int addrlen)
{
  int ret = connect(sd, addr, addrlen);

  // ret == -1 on failure
  if (ret == -1)
    perror("connect()");
  // ret == 0 on success

  return ret;
}

int Socket::sent(void *buf, int len, unsigned int s_flags, struct sockaddr *addr, int addrlen)
{
  int ret;
  
  if ((addr != NULL) && (addrlen > 0))
    ret = sendto(sd, buf, len, s_flags, addr, addrlen);
  else if (s_flags != 0)
    ret = send(sd, buf, len, s_flags);
  else {
    ret = write(sd, buf, len);
  }

  // ret == -1 on failure
/*
  if (ret == -1)
    perror("sendto()");
*/
  // ret >= 0 on success

  return ret;
}

int Socket::recw(void *buf, int len, unsigned int r_flags, struct sockaddr *addr, unsigned int *addrlen)
{
  int ret;

  if ((addr != NULL) && (addrlen != NULL))
    ret = recvfrom(sd, buf, len, r_flags, addr, addrlen);
  else if (r_flags != 0)
    ret = recv(sd, buf, len, r_flags);
  else
    ret = read(sd, buf, len);

  // ret == -1 on failure
  if (ret == -1)
    perror("recvfrom()");
  // ret >= 0 on success

  return ret;
}

int Socket::selectRead(struct timeval *timeout)
{
  fd_set rfds;

  FD_ZERO(&rfds); FD_SET(sd, &rfds);

  int ret = select(sd + 1, &rfds, NULL, NULL, timeout);

  // ret == -1 on failure
  if (ret == -1)
    perror("select()");
  // ret >= 0 on success

  return ret;
}

int Socket::selectWrite(struct timeval *timeout)
{
  fd_set wfds;

  FD_ZERO(&wfds); FD_SET(sd, &wfds);

  int ret = select(sd + 1, NULL, &wfds, NULL, timeout);

  // ret == -1 on failure
  if (ret == -1)
    perror("select()");
  // ret >= 0 on success

  return ret;
}

int Socket::selectExcept(struct timeval *timeout)
{
  fd_set efds;

  FD_ZERO(&efds); FD_SET(sd, &efds);

  int ret = select(sd + 1, NULL, NULL, &efds, timeout);

  // ret == -1 on failure
  if (ret == -1)
    perror("select()");
  // ret >= 0 on success

  return ret;
}


#include <netinet/in.h>


// TD: TOS-Feld setzen (IPv4)
void Socket::setTrafficClass(int trafficClass)
{
   int result = setsockopt(sd,SOL_IP,IP_TOS,&trafficClass,sizeof(int));
   if(result != 0) {
      printf("WARNING: Socket::setTrafficClass($%02x)!\n",trafficClass);
   }
}
