// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Synchronizable Implementation                                    ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "synchronizable.h"

#ifdef SYNC_DEBUG
#undef synchronized
#undef unsynchronized
#undef synchronizedTry
#undef resynchronize
#endif


#include <pthread.h>


namespace Coral {


// ###### Constructor #######################################################
Synchronizable::Synchronizable()
{
   // Initialize mutex for synchronization; mutextype has to be set to
   // PTHREAD_MUTEX_RECURSIVE to allow nested calls of synchronized() and
   // unsynchronized()!
   pthread_mutexattr_t mutexattr;
   pthread_mutexattr_init(&mutexattr);

   pthread_mutexattr_settype(&mutexattr,PTHREAD_MUTEX_RECURSIVE);
   // pthread_mutexattr_setkind_np(&mutexattr,PTHREAD_MUTEX_RECURSIVE_NP);

   pthread_mutex_init(&Mutex,&mutexattr);
}


// ###### Destructor ########################################################
Synchronizable::~Synchronizable()
{
   pthread_mutex_destroy(&Mutex);
}


// ###### Reinitialize ######################################################
void Synchronizable::resynchronize()
{
   pthread_mutex_destroy(&Mutex);
   pthread_mutexattr_t mutexattr;
   pthread_mutexattr_init(&mutexattr);
   pthread_mutexattr_settype(&mutexattr,PTHREAD_MUTEX_RECURSIVE_NP);
   pthread_mutex_init(&Mutex,&mutexattr);
}


// ###### Debug version of synchronized #####################################
void Synchronizable::synchronized_debug(const char* file, const cardinal line)
{
   cerr << "#" << getpid() << ": "
        << file << ", line " << line << ": " << "synchronize..." << endl;
   Synchronizable::synchronized();
   cerr << "#" << getpid() << ": "
        << file << ", line " << line << ": " << "synchronized!" << endl;
}


// ###### Debug version of unsynchronized ###################################
void Synchronizable::unsynchronized_debug(const char* file, const cardinal line)
{
   cerr << "#" << getpid() << ": "
        << file << ", line " << line << ": " << "unsynchronize..." << endl;
   Synchronizable::unsynchronized();
   cerr << "#" << getpid() << ": "
        << file << ", line " << line << ": " << "unsynchronized" << endl;
}


// ###### Debug version of synchronized #####################################
bool Synchronizable::synchronizedTry_debug(const char* file, const cardinal line)
{
   cerr << "#" << getpid() << ": "
        << file << ", line " << line << ": " << "try synchronize..." << endl;
   bool ok = Synchronizable::synchronizedTry();
   cerr << "#" << getpid() << ": " << file << ", line " << line << ": ";
   if(ok) {
      cerr << "synchronized!" << endl;
   }
   else {
      cerr << "locked -> not synchronized." << endl;
   }
   return(ok);
}


// ###### Debug version of resynchronize ####################################
void Synchronizable::resynchronize_debug(const char* file, const cardinal line)
{
   cerr << "#" << getpid() << ": "
        << file << ", line " << line << ": " << "Re-synchronize..." << endl;
   Synchronizable::resynchronize();
   cerr << "#" << getpid() << ": "
        << file << ", line " << line << ": " << "Re-synchronized!" << endl;
}


}
