// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Tools Implementation                                             ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include <sys/time.h>


#include "system.h"
#include "strings.h"
#include "tools.h"


namespace Coral {


// ###### Calculate the current time in milliseconds from 01-Jan-1970 #######
card64 getMicroTime()
{
  struct timeval tv;
  gettimeofday(&tv,NULL);
  return(((card64)tv.tv_sec * (card64)1000000) + (card64)tv.tv_usec);          
}


// ###### Calculate packets per second ######################################
cardinal calculatePacketsPerSecond(const cardinal payloadBytesPerSecond,
                                   const cardinal framesPerSecond,
                                   const cardinal maxPacketSize,
                                   const cardinal headerLength)
{
   const cardinal frameSize = (cardinal)ceil((double)payloadBytesPerSecond /
                                         (double)framesPerSecond);
   return((cardinal)ceil(frameSize / (double)(maxPacketSize - headerLength)) *
          (cardinal)ceil((double)payloadBytesPerSecond / (double)frameSize));
}


// ###### Calculate bytes per second ########################################
cardinal calculateBytesPerSecond(const cardinal payloadBytesPerSecond,
                                 const cardinal framesPerSecond,
                                 const cardinal maxPacketSize,
                                 const cardinal headerLength)
{
   const cardinal frameSize = (cardinal)ceil((double)payloadBytesPerSecond /
                                         (double)framesPerSecond);
   const cardinal headerBytesPerFrame =
      (cardinal)ceil(frameSize / (double)(maxPacketSize - headerLength)) *
      headerLength;
   
   return((headerBytesPerFrame + frameSize) *
            (cardinal)ceil((double)payloadBytesPerSecond / frameSize));
}


// ###### Scan URL ##########################################################
bool scanURL(const String& location,
             String&       protocol,
             String&       host,
             String&       path)
{
   String url = location;

   // ====== Get protocol ===================================================
   integer p1 = url.find("://");
   if(p1 < 0) {
      if(protocol.isNull()) {
         return(false);
      }
      p1 = 0;
   }
   else {
      protocol = url.left(p1);
      p1 += 3;
   }

   // ====== Get host =======================================================
   url = url.mid(p1);
   integer p2 = url.index('/');
   if(p2 < 0) {
      return(false);
   }
   host = url.left(p2);
   
   // ====== Get path =======================================================
   path = url.mid(p2 + 1);

   protocol = protocol.toLower();
   host     = host.toLower();
   return(true);
}


// ###### Print time stamp ##################################################
void printTimeStamp(ostream& os)
{
   char str[128];
   const card64 microTime = getMicroTime();
   const time_t timeStamp = microTime / 1000000;
   const struct tm *timeptr = localtime(&timeStamp);
   strftime((char*)&str,256,"%d-%b-%Y %H:%M:%S",timeptr);
   os << str;
   snprintf((char*)&str,sizeof(str),
            ".%04d: ",(cardinal)(microTime % 1000000) / 100);
   os << str;
}


}


#ifdef USE_EFENCE


#include "synchronizable.h"


// VERY IMPORTANT:
// malloc() and free() have to be synchronized in order to get correct
// results!!!
Coral::Synchronizable MAllocSync;


// ###### Operator new() replacement for libefence usage ####################
void* operator new(size_t size) throw (std::bad_alloc)
{
   MAllocSync.synchronized();
   if(size <= 0) {
      size = 1;
      cerr << "WARNING: operator new() - Size is 0!" << endl;
   }
   void* ptr = malloc(size);
   MAllocSync.unsynchronized();
   return(ptr);
}


// ###### Operator delete() replacement for libefence usage #################
void operator delete(void* ptr) throw ()
{
   MAllocSync.synchronized();
   free(ptr);
   MAllocSync.unsynchronized();
}


#endif
